@@ -2,7 +2,8 @@ from io import BytesIO
 from typing import IO
 from urllib.parse import urlparse
 
-from pyodide._core import IN_BROWSER
+from pyodide._package_loader import get_dynlibs, wheel_dist_info_dir
+from pyodide.ffi import IN_BROWSER, to_js
 from pyodide.http import pyfetch
 
 try:
@@ -41,4 +42,7 @@ __all__ = [
     "loadedPackages",
     "loadDynlib",
     "loadPackage",
+    "get_dynlibs",
+    "wheel_dist_info_dir",
+    "to_js",
 ]
