@@ -2,14 +2,19 @@ import asyncio
 import hashlib
 import importlib
 import json
+import shutil
+import site
+import sys
 import warnings
 from asyncio import gather
 from dataclasses import dataclass, field
 from importlib.metadata import PackageNotFoundError
+from importlib.metadata import distribution as importlib_distribution
 from importlib.metadata import distributions as importlib_distributions
 from importlib.metadata import version as importlib_version
 from pathlib import Path
 from sysconfig import get_platform
+from textwrap import dedent
 from typing import IO, Any
 from urllib.parse import ParseResult, urlparse
 from zipfile import ZipFile
@@ -19,17 +24,19 @@ from packaging.requirements import Requirement
 from packaging.tags import Tag, sys_tags
 from packaging.utils import canonicalize_name, parse_wheel_filename
 from packaging.version import Version
-from pyodide._package_loader import get_dynlibs, wheel_dist_info_dir
-from pyodide.ffi import to_js
 
+from . import _mock_package
 from ._compat import (
     REPODATA_INFO,
     REPODATA_PACKAGES,
     fetch_bytes,
     fetch_string,
+    get_dynlibs,
     loadDynlib,
     loadedPackages,
     loadPackage,
+    to_js,
+    wheel_dist_info_dir,
 )
 from .externals.pip._internal.utils.wheel import pkg_resources_distribution_for_wheel
 from .package import PackageDict, PackageMetadata
@@ -82,15 +89,28 @@ class WheelInfo:
             parsed_url=parsed_url,
         )
 
+    def best_compatible_tag_index(self) -> int | None:
+        """Get the index of the first tag in ``packaging.tags.sys_tags()`` that this wheel has.
+
+        Since ``packaging.tags.sys_tags()`` is sorted from most specific ("best") to most
+        general ("worst") compatibility, this index douples as a priority rank: given two
+        compatible wheels, the one whose best index is closer to zero should be installed.
+
+        Returns
+        -------
+        The index, or ``None`` if this wheel has no compatible tags.
+        """
+        for index, tag in enumerate(sys_tags()):
+            if tag in self.tags:
+                return index
+        return None
+
     def is_compatible(self):
         if self.filename.endswith("py3-none-any.whl"):
             return True
-        for tag in sys_tags():
-            if tag in self.tags:
-                return True
-        return False
+        return self.best_compatible_tag_index() is not None
 
-    def check_compatible(self):
+    def check_compatible(self) -> None:
         if self.is_compatible():
             return
         tag: Tag = next(iter(self.tags))
@@ -261,15 +281,23 @@ def find_wheel(metadata: dict[str, Any], req: Requirement) -> WheelInfo:
             )
             continue
 
+        best_wheel = None
+        best_tag_index = float("infinity")
+
         release = releases[str(ver)]
         for fileinfo in release:
             url = fileinfo["url"]
             if not url.endswith(".whl"):
                 continue
             wheel = WheelInfo.from_url(url)
-            if wheel.is_compatible():
+            tag_index = wheel.best_compatible_tag_index()
+            if tag_index is not None and tag_index < best_tag_index:
                 wheel.digests = fileinfo["digests"]
-                return wheel
+                best_wheel = wheel
+                best_tag_index = tag_index
+
+        if best_wheel is not None:
+            return wheel
 
     raise ValueError(
         f"Can't find a pure Python 3 wheel for '{req}'.\n"
@@ -307,7 +335,7 @@ class Transaction:
         if isinstance(req, Requirement):
             return await self.add_requirement_inner(req)
 
-        if not req.endswith(".whl"):
+        if not urlparse(req).path.endswith(".whl"):
             return await self.add_requirement_inner(Requirement(req))
 
         # custom download location
@@ -445,8 +473,6 @@ async def install(
 ) -> None:
     """Install the given package and all of its dependencies.
 
-    See :ref:`loading packages <loading_packages>` for more information.
-
     If a package is not found in the Pyodide repository it will be loaded from
     PyPI. Micropip can only load pure Python wheels or wasm32/emscripten wheels
     built by Pyodide.
@@ -457,7 +483,7 @@ async def install(
 
     Parameters
     ----------
-    requirements : ``str | List[str]``
+    requirements :
 
         A requirement or list of requirements to install. Each requirement is a
         string, which should be either a package name or a wheel URI:
@@ -473,7 +499,7 @@ async def install(
 
         - If a wheel URI starts with ``emfs:``, it will be interpreted as a path
           in the Emscripten file system (Pyodide's file system). E.g.,
-          `emfs:../relative/path/wheel.whl` or `emfs:/absolute/path/wheel.whl`.
+          ``emfs:../relative/path/wheel.whl`` or ``emfs:/absolute/path/wheel.whl``.
           In this case, only .whl files are supported.
 
         - If a wheel URI requirement starts with ``http:`` or ``https:`` it will
@@ -482,8 +508,7 @@ async def install(
         - In node, you can access the native file system using a URI that starts
           with ``file:``. In the browser this will not work.
 
-
-    keep_going : ``bool``, default: False
+    keep_going :
 
         This parameter decides the behavior of the micropip when it encounters a
         Python package without a pure Python wheel while doing dependency
@@ -495,12 +520,12 @@ async def install(
         - If ``True``, the micropip will keep going after the first error, and
           report a list of errors at the end.
 
-    deps : ``bool``, default: True
+    deps :
 
         If ``True``, install dependencies specified in METADATA file for each
         package. Otherwise do not install dependencies.
 
-    credentials : ``Optional[str]``
+    credentials :
 
         This parameter specifies the value of ``credentials`` when calling the
         `fetch() <https://developer.mozilla.org/en-US/docs/Web/API/fetch>`__
@@ -508,17 +533,11 @@ async def install(
 
         When not specified, ``fetch()`` is called without ``credentials``.
 
-    pre : ``bool``, default: False
+    pre :
 
         If ``True``, include pre-release and development versions. By default,
         micropip only finds stable versions.
 
-    Returns
-    -------
-    ``Future``
-
-        A ``Future`` that resolves to ``None`` when all packages have been
-        downloaded and installed.
     """
     ctx = default_environment()
     if isinstance(requirements, str):
@@ -588,12 +607,12 @@ def freeze() -> str:
     ``repodata.json`` lock file.
 
     If you later load Pyodide with this lock file, you can use
-    :any:`pyodide.loadPackage` to load packages that were loaded with `micropip`
-    this time. Loading packages with :any:`pyodide.loadPackage` is much faster
+    :js:func:`pyodide.loadPackage` to load packages that were loaded with :py:mod:`micropip`
+    this time. Loading packages with :js:func:`~pyodide.loadPackage` is much faster
     and you will always get consistent versions of all your dependencies.
 
     You can use your custom lock file by passing an appropriate url to the
-    `lockFileURL` argument to :any:`loadPyodide <globalThis.loadPyodide>`.
+    ``lockFileURL`` of :js:func:`~globalThis.loadPyodide`.
     """
     from copy import deepcopy
 
@@ -634,12 +653,12 @@ def freeze() -> str:
     return json.dumps(package_data)
 
 
-def _list():
+def _list() -> PackageDict:
     """Get the dictionary of installed packages.
 
     Returns
     -------
-    packages : :any:`micropip.package.PackageDict`
+    ``PackageDict``
         A dictionary of installed packages.
 
         >>> import micropip
@@ -691,3 +710,172 @@ def _list():
             source_ = pkg_source
         packages[name] = PackageMetadata(name=name, version=version, source=source_)
     return packages
+
+
+MOCK_INSTALL_NAME_MEMORY = "micropip in-memory mock package"
+MOCK_INSTALL_NAME_PERSISTENT = "micropip mock package"
+
+
+def add_mock_package(
+    name: str,
+    version: str,
+    *,
+    modules: dict[str, str | None] | None = None,
+    persistent: bool = False,
+) -> None:
+    """
+    Add a mock version of a package to the package dictionary.
+
+    This means that if it is a dependency, it is skipped on install.
+
+    By default a single empty module is installed with the same
+    name as the package. You can alternatively give one or more modules to make a
+    set of named modules.
+
+    The modules parameter is usually a dictionary mapping module name to module text.
+
+    .. code-block:: python
+
+        {
+            "mylovely_module":'''
+            def module_method(an_argument):
+                print("This becomes a module level argument")
+
+            module_value = "this value becomes a module level variable"
+            print("This is run on import of module")
+            '''
+        }
+
+    If you are adding the module in non-persistent mode, you can also pass functions
+    which are used to initialize the module on loading (as in `importlib.abc.loader.exec_module` ).
+    This allows you to do things like use `unittest.mock.MagicMock` classes for modules.
+
+    .. code-block:: python
+
+        def init_fn(module):
+            module.dict["WOO"]="hello"
+            print("Initing the module now!")
+
+        ...
+
+        {
+            "mylovely_module": init_fn
+        }
+
+    Parameters
+    ----------
+    name :
+
+        Package name to add
+
+    version :
+
+        Version of the package. This should be a semantic version string,
+        e.g. 1.2.3
+
+    modules :
+
+        Dictionary of module_name:string pairs.
+        The string contains the source of the mock module or is blank for
+        an empty module.
+
+    persistent :
+
+        If this is True, modules will be written to the file system, so they
+        persist between runs of python (assuming the file system persists).
+        If it is False, modules will be stored inside micropip in memory only.
+    """
+    if modules is None:
+        # make a single mock module with this name
+        modules = {name: ""}
+    # make the metadata
+    METADATA = f"""Metadata-Version: 1.1
+Name: {name}
+Version: {version}
+Summary: {name} mock package generated by micropip
+Author-email: {name}@micro.pip.non-working-fake-host
+"""
+    for n in modules.keys():
+        METADATA += f"Provides:{n}\n"
+
+    if persistent:
+        # make empty mock modules with the requested names in user site packages
+        site_packages = Path(site.getusersitepackages())
+        # in pyodide site packages isn't on sys.path initially
+        if not site_packages.exists():
+            site_packages.mkdir(parents=True, exist_ok=True)
+        if site_packages not in sys.path:
+            sys.path.append(str(site_packages))
+        metadata_dir = site_packages / (name + "-" + version + ".dist-info")
+        metadata_dir.mkdir(parents=True, exist_ok=False)
+        metadata_file = metadata_dir / "METADATA"
+        record_file = metadata_dir / "RECORD"
+        installer_file = metadata_dir / "INSTALLER"
+        file_list = []
+        file_list.append(metadata_file)
+        file_list.append(installer_file)
+        with open(metadata_file, "w") as mf:
+            mf.write(METADATA)
+        for n, s in modules.items():
+            if s is None:
+                s = ""
+            s = dedent(s)
+            path_parts = n.split(".")
+            dir_path = Path(site_packages, *path_parts)
+            dir_path.mkdir(exist_ok=True, parents=True)
+            init_file = dir_path / "__init__.py"
+            file_list.append(init_file)
+            with open(init_file, "w") as f:
+                f.write(s)
+        with open(installer_file, "w") as f:
+            f.write(MOCK_INSTALL_NAME_PERSISTENT)
+        with open(record_file, "w") as f:
+            for file in file_list:
+                f.write(f"{str(file)},,{file.stat().st_size}\n")
+            f.write(f"{str(record_file)},,\n")
+    else:
+        # make memory mocks of files
+        INSTALLER = MOCK_INSTALL_NAME_MEMORY
+        metafiles = {"METADATA": METADATA, "INSTALLER": INSTALLER}
+        _mock_package.add_in_memory_distribution(name, metafiles, modules)
+    importlib.invalidate_caches()
+
+
+def list_mock_packages() -> list[str]:
+    packages = []
+    for dist in importlib_distributions():
+        installer = dist.read_text("INSTALLER")
+        if installer is not None and (
+            installer == MOCK_INSTALL_NAME_PERSISTENT
+            or installer == MOCK_INSTALL_NAME_MEMORY
+        ):
+            packages.append(dist.name)
+    return packages
+
+
+def remove_mock_package(name: str) -> None:
+    d = importlib_distribution(name)
+    installer = d.read_text("INSTALLER")
+    if installer == MOCK_INSTALL_NAME_MEMORY:
+        _mock_package.remove_in_memory_distribution(name)
+        return
+    elif installer is None or installer != MOCK_INSTALL_NAME_PERSISTENT:
+        raise ValueError(
+            f"Package {name} doesn't seem to be a micropip mock. \n"
+            "Are you sure it was installed with micropip?"
+        )
+    # a real mock package - kill it
+    # remove all files
+    folders: set[Path] = set()
+    if d.files is not None:
+        for file in d.files:
+            p = Path(file.locate())
+            p.unlink()
+            folders.add(p.parent)
+    # delete all folders except site_packages
+    # (that check is just to avoid killing
+    # undesirable things in case of weird micropip errors)
+    site_packages = Path(site.getusersitepackages())
+    for f in folders:
+        if f != site_packages:
+            shutil.rmtree(f)
