@@ -2,16 +2,13 @@ import io
 import sys
 import zipfile
 from contextlib import contextmanager
+from importlib.metadata import Distribution, PackageNotFoundError
 from pathlib import Path
 from tempfile import TemporaryDirectory
 
 import pytest
 from pytest_pyodide import run_in_pyodide, spawn_web_server
 
-sys.path.append(str(Path(__file__).resolve().parent / "vendored"))
-
-from importlib.metadata import Distribution, PackageNotFoundError
-
 try:
     from packaging.tags import Tag
 
@@ -92,8 +89,12 @@ def make_wheel_filename(name: str, version: str, platform: str = "generic") -> s
         platform_str = "py3-none-any"
     elif platform == "emscripten":
         platform_str = f"{cpver}-{cpver}-{_platform()}"
-    elif platform == "native":
+    elif platform == "linux":
         platform_str = f"{cpver}-{cpver}-manylinux_2_31_x86_64"
+    elif platform == "windows":
+        platform_str = f"{cpver}-{cpver}-win_amd64"
+    elif platform == "invalid":
+        platform_str = f"{cpver}-{cpver}-invalid"
     else:
         platform_str = platform
 
@@ -194,17 +195,48 @@ def mock_fetch(monkeypatch, mock_importlib, wheel_base):
     return result
 
 
+@pytest.fixture(scope="module")
+def wheel_path(tmp_path_factory):
+    # Build a micropip wheel for testing
+    import build
+    from build.env import IsolatedEnvBuilder
+
+    output_dir = tmp_path_factory.mktemp("wheel")
+
+    with IsolatedEnvBuilder() as env:
+        builder = build.ProjectBuilder(Path(__file__).parent.parent)
+        builder.python_executable = env.executable
+        builder.scripts_dir = env.scripts_dir
+        env.install(builder.build_system_requires)
+        builder.build("wheel", output_directory=output_dir)
+
+    yield output_dir
+
+
 @pytest.fixture
-def selenium_standalone_micropip(selenium_standalone):
+def selenium_standalone_micropip(selenium_standalone, wheel_path):
     """Import micropip before entering test so that global initialization of
     micropip doesn't count towards hiwire refcount.
     """
-    selenium_standalone.run_js(
-        """
-        await pyodide.loadPackage("micropip");
-        pyodide.runPython("import micropip");
-        """
-    )
+
+    wheel_dir = Path(wheel_path)
+    wheel_files = list(wheel_dir.glob("*.whl"))
+
+    if not wheel_files:
+        pytest.exit("No wheel files found in wheel/ directory")
+
+    wheel_file = wheel_files[0]
+    with spawn_web_server(wheel_dir) as server:
+        server_hostname, server_port, _ = server
+        base_url = f"http://{server_hostname}:{server_port}/"
+        selenium_standalone.run_js(
+            f"""
+            await pyodide.loadPackage("{base_url + wheel_file.name}");
+            await pyodide.loadPackage(["packaging", "pyparsing"]);
+            pyodide.runPython("import micropip");
+            """
+        )
+
     yield selenium_standalone
 
 
@@ -388,6 +420,23 @@ async def test_add_requirement_marker(mock_importlib, wheel_base):
         assert t not in wheel_files
 
 
+@pytest.mark.asyncio
+async def test_add_requirement_query_url(mock_importlib, wheel_base, monkeypatch):
+    pytest.importorskip("packaging")
+    from micropip._micropip import Transaction
+
+    async def mock_add_wheel(self, wheel, extras):
+        self.mock_wheel = wheel
+
+    monkeypatch.setattr(Transaction, "add_wheel", mock_add_wheel)
+
+    transaction = create_transaction(Transaction)
+    await transaction.add_requirement(f"{SNOWBALL_WHEEL}?b=1")
+    wheel = transaction.mock_wheel
+    assert wheel.name == "snowballstemmer"
+    assert wheel.filename == SNOWBALL_WHEEL  # without the query params
+
+
 @pytest.mark.asyncio
 async def test_package_with_extra(mock_fetch):
     mock_fetch.add_pkg_version("depa")
@@ -449,6 +498,30 @@ async def test_package_with_extra_transitive(
         assert "depb" not in pkg_list
 
 
+def _pypi_metadata(package, versions_to_tags):
+    # Build package release metadata as would be returned from
+    # https://pypi.org/pypi/{pkgname}/json
+    #
+    # `package` is a string containing the package name as
+    # it would appear in a wheel file name.
+    #
+    # `versions` is a mapping with version strings as
+    # keys and iterables of tag strings as values.
+    releases = {}
+    for version, tags in versions_to_tags.items():
+        release = []
+        for tag in tags:
+            wheel_name = f"{package}-{version}-{tag}-none-any.whl"
+            wheel_info = {
+                "filename": wheel_name,
+                "url": wheel_name,
+                "digests": None,
+            }
+            release.append(wheel_info)
+        releases[version] = release
+    return {"releases": releases}
+
+
 def test_last_version_from_pypi():
     pytest.importorskip("packaging")
     from packaging.requirements import Requirement
@@ -458,14 +531,7 @@ def test_last_version_from_pypi():
     requirement = Requirement("dummy_module")
     versions = ["0.0.1", "0.15.5", "0.9.1"]
 
-    # building metadata as returned from
-    # https://pypi.org/pypi/{pkgname}/json
-    releases = {}
-    for v in versions:
-        filename = f"dummy_module-{v}-py3-none-any.whl"
-        releases[v] = [{"filename": filename, "url": filename, "digests": None}]
-
-    metadata = {"releases": releases}
+    metadata = _pypi_metadata("dummy_module", {v: ["py3"] for v in versions})
 
     # get version number from find_wheel
     wheel = find_wheel(metadata, requirement)
@@ -473,6 +539,78 @@ def test_last_version_from_pypi():
     assert str(wheel.version) == "0.15.5"
 
 
+_best_tag_test_cases = (
+    "package, version, incompatible_tags, compatible_tags",
+    # Tests assume that `compatible_tags` is sorted from least to most compatible:
+    [
+        # Common modern case (pure Python 3-only wheel):
+        ("hypothesis", "6.60.0", [], ["py3"]),
+        # Common historical case (pure Python 2-or-3 wheel):
+        ("attrs", "22.1.0", [], ["py2.py3"]),
+        # Still simple, less common (separate Python 2 and 3 wheels):
+        ("raise", "1.1.9", ["py2"], ["py3"]),
+        # More complicated, rarer cases:
+        ("compose", "1.4.8", [], ["py2.py30", "py35", "py38"]),
+        ("with_as_a_function", "1.0.1", ["py20", "py25"], ["py26.py3"]),
+        ("with_as_a_function", "1.1.0", ["py22", "py25"], ["py26.py30", "py33"]),
+    ],
+)
+
+
+@pytest.mark.parametrize(*_best_tag_test_cases)
+def test_best_tag_from_pypi(package, version, incompatible_tags, compatible_tags):
+    pytest.importorskip("packaging")
+    from packaging.requirements import Requirement
+
+    from micropip._micropip import find_wheel
+
+    requirement = Requirement(package)
+    tags = incompatible_tags + compatible_tags
+
+    metadata = _pypi_metadata(package, {version: tags})
+
+    wheel = find_wheel(metadata, requirement)
+
+    best_tag = tags[-1].split(".")[-1] + "-none-any"
+    assert best_tag in set(map(str, wheel.tags))
+
+
+# A newer version with a compatible wheel has higher precedence
+# than an older version with a more precisely compatible wheel.
+# This test verifies that we didn't break that corner case:
+@pytest.mark.parametrize(
+    "package, old_version, old_tags, new_version, new_tags",
+    [
+        ("compose", "1.1.1", ["py2.py3"], "1.2.0", ["py2.py30", "py35", "py38"]),
+        (
+            "with_as_a_function",
+            "1.0.1",
+            ["py20", "py25", "py26.py3"],
+            "1.1.0",
+            ["py22", "py25", "py26.py30", "py33"],
+        ),
+    ],
+)
+def test_last_version_and_best_tag_from_pypi(
+    package, old_version, new_version, old_tags, new_tags
+):
+    pytest.importorskip("packaging")
+    from packaging.requirements import Requirement
+
+    from micropip._micropip import find_wheel
+
+    requirement = Requirement(package)
+
+    metadata = _pypi_metadata(
+        package,
+        {old_version: old_tags, new_version: new_tags},
+    )
+
+    wheel = find_wheel(metadata, requirement)
+
+    assert str(wheel.version) == new_version
+
+
 @pytest.mark.asyncio
 async def test_install_non_pure_python_wheel():
     pytest.importorskip("packaging")
@@ -544,15 +682,14 @@ def test_install_mixed_case2(selenium_standalone_micropip, jinja2):
     )
 
 
-@pytest.mark.xfail(reason="test fails in some environments")
 @pytest.mark.asyncio
 async def test_install_keep_going(mock_fetch: mock_fetch_cls) -> None:
     dummy = "dummy"
     dep1 = "dep1"
     dep2 = "dep2"
     mock_fetch.add_pkg_version(dummy, requirements=[dep1, dep2])
-    mock_fetch.add_pkg_version(dep1, platform="native")
-    mock_fetch.add_pkg_version(dep2, platform="native")
+    mock_fetch.add_pkg_version(dep1, platform="invalid")
+    mock_fetch.add_pkg_version(dep2, platform="invalid")
 
     # report order is non-deterministic
     msg = f"({dep1}|{dep2}).*({dep2}|{dep1})"
@@ -903,3 +1040,191 @@ def test_check_compatible(mock_platform, interp, abi, arch, ctx):
     wheel_name = f"{pkg}-{interp}-{abi}-{arch}.whl"
     with ctx:
         WheelInfo.from_url(wheel_name).check_compatible()
+
+
+@pytest.mark.parametrize(*_best_tag_test_cases)
+def test_best_compatible_tag(package, version, incompatible_tags, compatible_tags):
+    from micropip._micropip import WheelInfo
+
+    for tag in incompatible_tags:
+        wheel_name = f"{package}-{version}-{tag}-none-any.whl"
+        wheel = WheelInfo.from_url(wheel_name)
+        assert wheel.best_compatible_tag_index() is None
+
+    wheels = []
+    for tag in compatible_tags:
+        wheel_name = f"{package}-{version}-{tag}-none-any.whl"
+        wheel = WheelInfo.from_url(wheel_name)
+        wheels.append(wheel)
+
+    sorted_wheels = sorted(wheels, key=WheelInfo.best_compatible_tag_index)
+    sorted_wheels.reverse()
+    assert sorted_wheels == wheels
+
+
+@run_in_pyodide()
+def test_persistent_mock_pyodide(selenium_standalone_micropip):  #
+    import sys
+    from importlib.metadata import version as importlib_version
+    from io import StringIO
+
+    from micropip._micropip import add_mock_package
+
+    capture_stdio = StringIO()
+
+    sys.stdout = capture_stdio
+    add_mock_package("test_1", "1.0.0", persistent=True)
+    add_mock_package(
+        "test_2",
+        "1.2.0",
+        modules={
+            "t1": "print('hi from t1')",
+            "t2": """
+        def fn():
+            print("Hello from fn")
+        """,
+        },
+        persistent=True,
+    )
+    import test_1
+
+    dir(test_1)
+
+    import t1
+
+    dir(t1)
+    import t2
+
+    dir(t2)
+
+    t2.fn()
+    assert capture_stdio.getvalue().find("Hello from fn") != -1
+    assert capture_stdio.getvalue().find("hi from t1") != -1
+    assert importlib_version("test_1") == "1.0.0"
+
+
+def test_persistent_mock(monkeypatch, capsys, tmp_path):
+    import site
+    from importlib.metadata import version as importlib_version
+
+    from micropip._micropip import add_mock_package
+
+    def _getusersitepackages():
+        return tmp_path
+
+    monkeypatch.setattr(site, "getusersitepackages", _getusersitepackages)
+    monkeypatch.setattr(sys, "path", [tmp_path])
+    add_mock_package("test_1", "1.0.0")
+    add_mock_package(
+        "test_2",
+        "1.2.0",
+        modules={
+            "t1": "print('hi from t1')",
+            "t2": """
+        def fn():
+            print("Hello from fn")
+        """,
+        },
+    )
+    import t1
+
+    dir(t1)
+    import t2
+
+    dir(t2)
+    import test_1
+
+    dir(test_1)
+
+    t2.fn()
+    assert importlib_version("test_2") == "1.2.0"
+    captured = capsys.readouterr()
+    assert captured.out.find("hi from t1") != -1
+    assert captured.out.find("Hello from fn") != -1
+
+
+def test_memory_mock():
+    from micropip import _micropip
+
+    def mod_init(module):
+        module.__dict__["add2"] = lambda x: x + 2
+
+    _micropip.add_mock_package(
+        "micropip_test_bob",
+        "1.0.0",
+        modules={
+            "micropip_bob_mod": "print('hi from bob')",
+            "micropip_bob_mod.fn": mod_init,
+        },
+        persistent=False,
+    )
+    import importlib
+
+    import micropip_bob_mod
+
+    dir(micropip_bob_mod)
+
+    import micropip_bob_mod.fn
+
+    assert micropip_bob_mod.fn.add2(5) == 7
+
+    found_bob = False
+    for d in importlib.metadata.distributions():
+        if d.name == "micropip_test_bob":
+            found_bob = True
+    assert found_bob is True
+    assert (
+        importlib.metadata.distribution("micropip_test_bob").name == "micropip_test_bob"
+    )
+    # check package removes okay
+    _micropip.remove_mock_package("micropip_test_bob")
+    del micropip_bob_mod
+    with pytest.raises(ImportError):
+        import micropip_bob_mod
+
+        dir(micropip_bob_mod)
+
+
+@run_in_pyodide()
+def test_memory_mock_pyodide(selenium_standalone_micropip):
+    import pytest
+
+    from micropip import _micropip
+
+    def mod_init(module):
+        module.__dict__["add2"] = lambda x: x + 2
+
+    _micropip.add_mock_package(
+        "micropip_test_bob",
+        "1.0.0",
+        modules={
+            "micropip_bob_mod": "print('hi from bob')",
+            "micropip_bob_mod.fn": mod_init,
+        },
+        persistent=False,
+    )
+    import importlib
+
+    import micropip_bob_mod
+
+    dir(micropip_bob_mod)
+
+    import micropip_bob_mod.fn
+
+    assert micropip_bob_mod.fn.add2(5) == 7
+
+    found_bob = False
+    for d in importlib.metadata.distributions():
+        if d.name == "micropip_test_bob":
+            found_bob = True
+    assert found_bob is True
+    assert (
+        importlib.metadata.distribution("micropip_test_bob").name == "micropip_test_bob"
+    )
+    # check package removes okay
+    _micropip.remove_mock_package("micropip_test_bob")
+    del micropip_bob_mod
+    with pytest.raises(ImportError):
+        import micropip_bob_mod
+
+        dir(micropip_bob_mod)
